import threading

import pytest
import sqlalchemy as sa


@pytest.fixture()
def mgr(monkeypatch, tmp_path):
    """
    Provide a configured singleton Database with isolated state:
    - reset internal state between tests
    - force config to use tmp_path as ostryalis.directory
    """
    from ostryalis.database import Database

    m = Database()
    m.reset()
    m.set_config_getter(lambda key: str(tmp_path))

    yield m

    m.reset()


def _attached_aliases_via_pragma(dbapi_conn):
    cur = dbapi_conn.cursor()
    try:
        cur.execute("PRAGMA database_list")
        return {row[1] for row in cur.fetchall()}
    finally:
        cur.close()


def test_validate_alias_rejects_empty(mgr):
    with pytest.raises(ValueError):
        mgr._validate_alias("")
    with pytest.raises(ValueError):
        mgr._validate_alias("   ")


def test_validate_alias_rejects_reserved(mgr):
    with pytest.raises(ValueError):
        mgr._validate_alias("main")
    with pytest.raises(ValueError):
        mgr._validate_alias("temp")


def test_validate_alias_rejects_invalid_identifier(mgr):
    with pytest.raises(ValueError):
        mgr._validate_alias("1abc")
    with pytest.raises(ValueError):
        mgr._validate_alias("foo-bar")
    with pytest.raises(ValueError):
        mgr._validate_alias("foo bar")


def test_validate_alias_normalizes_whitespace(mgr):
    assert mgr._validate_alias("  abc_123  ") == "abc_123"


def test_register_required_aliases_creates_dir_and_paths(mgr, tmp_path):
    aliases = [mgr.attach("a"), mgr.attach("b_2")]
    assert aliases == ["a", "b_2"]

    assert tmp_path.exists()
    assert mgr._attachments["a"] == str(tmp_path / "a.db")
    assert mgr._attachments["b_2"] == str(tmp_path / "b_2.db")


def test_get_engine_creates_main_db_and_session_factory(mgr, tmp_path):
    engine = mgr.engine
    assert engine is not None

    expected = tmp_path / "main.db"
    with engine.connect():
        pass
    assert expected.exists()


def test_sqlite_on_connect_attaches_known_aliases(mgr):
    mgr.attach("alpha")
    mgr.attach("beta")
    engine = mgr.engine

    with engine.connect() as conn:
        dbapi = conn.connection
        names = _attached_aliases_via_pragma(dbapi)
        assert "alpha" in names
        assert "beta" in names


def test_ensure_session_commit_true_persists_changes(mgr):
    engine = mgr.engine

    with mgr.session_scope(commit=True) as s:
        s.execute(sa.text("CREATE TABLE IF NOT EXISTS t (id INTEGER PRIMARY KEY, v TEXT)"))
        s.execute(sa.text("DELETE FROM t"))
        s.execute(sa.text("INSERT INTO t (v) VALUES ('x')"))

    with engine.connect() as c:
        row = c.execute(sa.text("SELECT COUNT(*) FROM t")).scalar_one()
        assert row == 1


def test_ensure_session_commit_false_does_not_commit(mgr):
    engine = mgr.engine
    with engine.begin() as c:
        c.execute(sa.text("CREATE TABLE IF NOT EXISTS t2 (id INTEGER PRIMARY KEY, v TEXT)"))
        c.execute(sa.text("DELETE FROM t2"))

    with mgr.session_scope(commit=False) as s:
        s.execute(sa.text("INSERT INTO t2 (v) VALUES ('y')"))

    with engine.connect() as c:
        row = c.execute(sa.text("SELECT COUNT(*) FROM t2")).scalar_one()
        assert row == 0


def test_ensure_session_rollback_on_exception(mgr):
    engine = mgr.engine
    with engine.begin() as c:
        c.execute(sa.text("CREATE TABLE IF NOT EXISTS t3 (id INTEGER PRIMARY KEY, v TEXT)"))
        c.execute(sa.text("DELETE FROM t3"))

    with pytest.raises(RuntimeError):
        with mgr.session_scope(commit=True) as s:
            s.execute(sa.text("INSERT INTO t3 (v) VALUES ('z')"))
            raise RuntimeError("boom")

    with engine.connect() as c:
        row = c.execute(sa.text("SELECT COUNT(*) FROM t3")).scalar_one()
        assert row == 0


def test_ensure_session_with_existing_session_does_not_close_it(mgr):
    factory = mgr._get_session_factory()
    s = factory()
    try:
        with mgr.session_scope(s, require_attached=["xalias"]) as s2:
            assert s2 is s
            sa_conn = s2.connection()
            names = _attached_aliases_via_pragma(sa_conn.connection)
            assert "xalias" in names

        s.execute(sa.text("SELECT 1")).scalar_one()
    finally:
        s.close()


def test_attachment_cache_skips_repeated_pragma_calls(mgr, monkeypatch):
    _ = mgr.engine

    real = mgr._sqlite_attach_aliases
    pragma_counter = {"count": 0}

    def wrapped(dbapi_connection, *, aliases, connection_record=None):
        if not hasattr(dbapi_connection, "set_trace_callback"):
            pytest.skip("DB-API connection does not support set_trace_callback; cannot count PRAGMA.")

        def trace_cb(stmt: str):
            if stmt.strip().lower() == "pragma database_list":
                pragma_counter["count"] += 1

        prev = dbapi_connection.set_trace_callback(trace_cb)
        try:
            return real(dbapi_connection, aliases=aliases, connection_record=connection_record)
        finally:
            dbapi_connection.set_trace_callback(prev)

    monkeypatch.setattr(mgr, "_sqlite_attach_aliases", wrapped)

    with mgr.session_scope(require_attached=["cachetest"]) as s:
        sa_conn = s.connection()
        names = _attached_aliases_via_pragma(sa_conn.connection)
        assert "cachetest" in names

    first = pragma_counter["count"]
    assert first >= 1

    with mgr.session_scope(require_attached=["cachetest"]) as s:
        sa_conn = s.connection()
        names = _attached_aliases_via_pragma(sa_conn.connection)
        assert "cachetest" in names

    second = pragma_counter["count"]
    assert second == first


def test_attachment_cache_is_per_connection(mgr, monkeypatch, tmp_path):
    engine = mgr.engine

    pragma_counter = {"count": 0}
    real = mgr._sqlite_attach_aliases

    def traced(dbapi_connection, *, aliases, connection_record=None):
        cb_count = {"n": 0}

        def trace_cb(stmt):
            if stmt.strip().lower() == "pragma database_list":
                cb_count["n"] += 1

        prev = None
        if hasattr(dbapi_connection, "set_trace_callback"):
            prev = dbapi_connection.set_trace_callback(trace_cb)
        try:
            return real(dbapi_connection, aliases=aliases, connection_record=connection_record)
        finally:
            if hasattr(dbapi_connection, "set_trace_callback"):
                dbapi_connection.set_trace_callback(prev)
            pragma_counter["count"] += cb_count["n"]

    monkeypatch.setattr(mgr, "_sqlite_attach_aliases", traced)

    with mgr.session_scope(require_attached=["perconn"]) as s:
        pass

    first = pragma_counter["count"]
    assert first >= 1

    engine.dispose()
    mgr.reset()
    mgr.set_config_getter(lambda key: str(tmp_path))
    _ = mgr.engine

    with mgr.session_scope(require_attached=["perconn"]) as s:
        pass

    second = pragma_counter["count"]
    assert second > first


def test_concurrency_register_required_aliases_is_thread_safe(mgr):
    names = [f"a{i}" for i in range(20)]
    errors = []
    barrier = threading.Barrier(parties=10)

    def worker(tid: int):
        try:
            barrier.wait()
            local = names[tid % 5 : tid % 5 + 10]
            for a in local:
                mgr.attach(a)
        except Exception as e:
            errors.append(e)

    threads = [threading.Thread(target=worker, args=(i,)) for i in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()

    assert not errors, f"Unexpected errors: {errors!r}"

    for alias, path in mgr._attachments.items():
        assert path.endswith(f"{alias}.db")


def test_concurrency_ensure_session_require_attached_no_exceptions(mgr):
    _ = mgr.engine
    mgr.attach("conc")

    errors = []
    barrier = threading.Barrier(parties=10)

    def worker():
        try:
            barrier.wait()
            with mgr.session_scope(require_attached=["conc"], commit=False) as s:
                s.execute(sa.text("SELECT 1")).scalar_one()
                sa_conn = s.connection()
                cur = sa_conn.connection.cursor()
                try:
                    cur.execute("PRAGMA database_list")
                    names = {row[1] for row in cur.fetchall()}
                finally:
                    cur.close()
                assert "conc" in names
        except Exception as e:
            errors.append(e)

    threads = [threading.Thread(target=worker) for _ in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()

    assert not errors, f"Unexpected errors: {errors!r}"


def test_concurrency_mixed_alias_registration_and_attachment(mgr):
    _ = mgr.engine

    aliases = [f"m{i}" for i in range(10)]
    errors = []
    barrier = threading.Barrier(parties=20)

    def registrar(i: int):
        try:
            barrier.wait()
            mgr.attach(aliases[i])
        except Exception as e:
            errors.append(e)

    def attacher(i: int):
        try:
            barrier.wait()
            with mgr.session_scope(require_attached=[aliases[i]], commit=False) as s:
                s.execute(sa.text("SELECT 1")).scalar_one()
        except Exception as e:
            errors.append(e)

    threads = []
    for i in range(10):
        threads.append(threading.Thread(target=registrar, args=(i,)))
        threads.append(threading.Thread(target=attacher, args=(i,)))

    for t in threads:
        t.start()
    for t in threads:
        t.join()

    assert not errors, f"Unexpected errors: {errors!r}"
    for a in aliases:
        assert a in mgr._attachments


# -----------------------
# New tests: SQLite UDFs
# -----------------------

def test_validate_udf_name_rejects_empty(mgr):
    with pytest.raises(ValueError):
        mgr._validate_udf_name("")
    with pytest.raises(ValueError):
        mgr._validate_udf_name("   ")


def test_validate_udf_name_rejects_invalid_identifier(mgr):
    with pytest.raises(ValueError):
        mgr._validate_udf_name("1abc")
    with pytest.raises(ValueError):
        mgr._validate_udf_name("foo-bar")
    with pytest.raises(ValueError):
        mgr._validate_udf_name("foo bar")


def test_validate_udf_arity_rejects_negative(mgr):
    with pytest.raises(ValueError):
        mgr._validate_udf_arity(-1)


def test_register_function_and_require_functions_installs_and_executes(mgr):
    _ = mgr.engine

    def add1(x):
        return x + 1

    mgr.register_function("add1", 1, add1)

    with mgr.session_scope(require_functions=[("add1", 1)], commit=False) as s:
        v = s.execute(sa.text("SELECT add1(41)")).scalar_one()
        assert v == 42


def test_udf_overload_by_arity_works(mgr):
    _ = mgr.engine

    def f1(x):
        return x + 10

    def f2(x, y):
        return x + y + 100

    mgr.register_function("ovl", 1, f1)
    mgr.register_function("ovl", 2, f2)

    with mgr.session_scope(require_functions=[("ovl", 1), ("ovl", 2)], commit=False) as s:
        a = s.execute(sa.text("SELECT ovl(5)")).scalar_one()
        b = s.execute(sa.text("SELECT ovl(5, 7)")).scalar_one()
        assert a == 15
        assert b == 112


def test_sqlite_on_connect_installs_registered_udfs(mgr):
    # Register before engine creates first connection; connect event should install automatically.
    def mul2(x):
        return x * 2

    mgr.register_function("mul2", 1, mul2)
    engine = mgr.engine

    with engine.connect() as c:
        v = c.execute(sa.text("SELECT mul2(21)")).scalar_one()
        assert v == 42


def test_require_functions_missing_registration_raises_keyerror(mgr):
    _ = mgr.engine
    with pytest.raises(KeyError):
        with mgr.session_scope(require_functions=[("nope", 1)], commit=False) as s:
            pass


def test_udf_cache_is_per_connection(mgr, monkeypatch, tmp_path):
    engine = mgr.engine

    def inc(x):
        return x + 1

    mgr.register_function("inc", 1, inc)

    real = mgr._sqlite_install_udfs
    counter = {"count": 0}

    def wrapped(dbapi_connection, *, required, connection_record=None):
        counter["count"] += 1
        return real(dbapi_connection, required=required, connection_record=connection_record)

    monkeypatch.setattr(mgr, "_sqlite_install_udfs", wrapped)

    with mgr.session_scope(require_functions=[("inc", 1)], commit=False) as s:
        pass

    first = counter["count"]
    assert first >= 1

    engine.dispose()
    mgr.reset()
    mgr.set_config_getter(lambda key: str(tmp_path))
    _ = mgr.engine
    mgr.register_function("inc", 1, inc)

    with mgr.session_scope(require_functions=[("inc", 1)], commit=False) as s:
        pass

    second = counter["count"]
    assert second > first


def test_concurrency_register_function_is_thread_safe(mgr):
    _ = mgr.engine
    errors = []
    barrier = threading.Barrier(parties=10)

    def make_func(k):
        return lambda x, _k=k: x + _k

    def worker(i: int):
        try:
            barrier.wait()
            mgr.register_function("cf", i % 3, make_func(i))
        except Exception as e:
            errors.append(e)

    threads = [threading.Thread(target=worker, args=(i,)) for i in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()

    assert not errors, f"Unexpected errors: {errors!r}"
    # Should have at most 3 overloads due to setdefault semantics.
    assert len([k for k in mgr._udfs.keys() if k[0] == "cf"]) <= 3


def test_concurrency_ensure_session_require_functions_no_exceptions(mgr):
    _ = mgr.engine

    def inc(x):
        return x + 1

    mgr.register_function("inc", 1, inc)

    errors = []
    barrier = threading.Barrier(parties=10)

    def worker():
        try:
            barrier.wait()
            with mgr.session_scope(require_functions=[("inc", 1)], commit=False) as s:
                assert s.execute(sa.text("SELECT inc(41)")).scalar_one() == 42
        except Exception as e:
            errors.append(e)

    threads = [threading.Thread(target=worker) for _ in range(10)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()

    assert not errors, f"Unexpected errors: {errors!r}"



__all__ = ["Database"]

import re
import threading
from contextlib import contextmanager
from pathlib import Path
from typing import Iterable, Callable, Any

from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker, Session

from inceptum import config


_ALIAS_RE = re.compile(r"^[A-Za-z_][A-Za-z0-9_]*$")
_RESERVED_ALIASES = {"main", "temp"}

# Key used in SQLAlchemy ConnectionRecord.info
_ATTACHED_CACHE_KEY = "ostryalis_sqlite_attached_aliases"

# Key used in SQLAlchemy ConnectionRecord.info
_UDF_CACHE_KEY = "ostryalis_sqlite_installed_udfs"

# Name validation for UDFs: allow dotted names? In SQLite, function names are identifiers;
# keep it conservative and consistent with aliases.
_UDF_NAME_RE = _ALIAS_RE


class Database:
    """
    Singleton-style database façade that owns all session/engine/attachment/UDF state.

    Construction is idempotent: Database() always returns the same instance.
    """

    _singleton_lock = threading.RLock()
    _singleton: "Database | None" = None

    def __new__(cls, *args, **kwargs):
        with cls._singleton_lock:
            if cls._singleton is None:
                cls._singleton = super().__new__(cls)
                cls._singleton._initialized = False
            return cls._singleton

    def __init__(self):
        if getattr(self, "_initialized", False):
            return
        self._initialized = True

        self._config = config

        self._engine: Engine | None = None
        self._SessionLocal: sessionmaker | None = None

        # alias -> absolute path string
        self._attachments: dict[str, str] = {}

        # (name, arity) -> python callable
        self._udfs: dict[tuple[str, int], Callable[..., Any]] = {}

        self._lock = threading.RLock()

    # Public API

    @property
    def engine(self) -> Engine:
        return self._get_engine()

    @contextmanager
    def session_scope(
        self,
        session: Session | None = None,
        *,
        commit: bool = True,
        require_attached: Iterable[str] | None = None,
        require_functions: Iterable[tuple[str, int]] | None = None,
    ):
        """
        If `session` is provided, it is reused and never committed/rolled back/closed
        by this context manager.

        If `session` is not provided, a new session is created; it is committed if
        `commit=True`, rolled back on exception, and always closed.
        """
        if session is not None:
            self._ensure_required_attached(session, require_attached)
            self._ensure_required_functions(session, require_functions)
            yield session
            return

        factory = self._get_session_factory()
        new_session = factory()
        try:
            self._ensure_required_attached(new_session, require_attached)
            self._ensure_required_functions(new_session, require_functions)
            yield new_session
            if commit:
                new_session.commit()
        except Exception:
            new_session.rollback()
            raise
        finally:
            new_session.close()

    def attach(self, alias: str, path: str | Path | None = None) -> str:
        """
        Register an attachment alias.

        If `path` is not provided, uses `{base_dir}/{alias}.db`.

        This only registers the mapping; actual ATTACH happens on connect
        and when `require_attached` is used.
        """
        alias = self._validate_alias(alias)
        if path is None:
            db_path = self._base_dir() / f"{alias}.db"
        else:
            db_path = Path(path).expanduser()
            if not db_path.is_absolute():
                # Relative paths become relative to base dir for predictability.
                db_path = self._base_dir() / db_path

        # Do not create the file here; ATTACH will create it if needed.
        db_path.parent.mkdir(parents=True, exist_ok=True)

        with self._lock:
            self._attachments.setdefault(alias, str(db_path))
        return alias

    def register_function(
        self,
        name: str,
        arity: int,
        func: Callable[..., Any],
    ) -> tuple[str, int]:
        """
        Register a SQLite user-defined function (UDF) by (name, arity).

        Registration is manager-level only; installation happens on engine connect
        and when `require_functions` is used.
        """
        name = self._validate_udf_name(name)
        arity = self._validate_udf_arity(arity)
        if not callable(func):
            raise ValueError("func must be callable.")

        key = (name, arity)
        with self._lock:
            # Like attachments, be idempotent: first one wins unless you explicitly change it.
            self._udfs.setdefault(key, func)
        return key

    # Private helpers

    def _get_engine(self) -> Engine:
        with self._lock:
            if self._engine is not None:
                return self._engine

            base_dir = self._base_dir()
            db_url = f"sqlite:///{base_dir / 'main.db'}"

            self._engine = create_engine(
                db_url,
                connect_args={"check_same_thread": False},
                future=True,
            )
            event.listen(self._engine, "connect", self._sqlite_on_connect)

            self._SessionLocal = sessionmaker(
                bind=self._engine,
                autoflush=False,
                autocommit=False,
                expire_on_commit=False,
                future=True,
            )
            return self._engine

    def _validate_alias(self, alias: str) -> str:
        if not isinstance(alias, str) or not alias.strip():
            raise ValueError("Database alias must be a non-empty string.")
        alias = alias.strip()
        if alias in _RESERVED_ALIASES:
            raise ValueError(f"Database alias '{alias}' is reserved by SQLite.")
        if not _ALIAS_RE.match(alias):
            raise ValueError("Database alias must match ^[A-Za-z_][A-Za-z0-9_]*$.")
        return alias

    def _validate_udf_name(self, name: str) -> str:
        if not isinstance(name, str) or not name.strip():
            raise ValueError("Function name must be a non-empty string.")
        name = name.strip()
        if not _UDF_NAME_RE.match(name):
            raise ValueError("Function name must match ^[A-Za-z_][A-Za-z0-9_]*$.")
        return name

    def _validate_udf_arity(self, arity: int) -> int:
        if not isinstance(arity, int):
            raise ValueError("Function arity must be an integer.")
        # sqlite3.create_function supports -1 for varargs, but that collides with the
        # “overload by arity” requirement in confusing ways
        if arity < 0:
            raise ValueError("Function arity must be >= 0.")
        return arity

    def _base_dir(self) -> Path:
        base_dir = Path(str(self._config("ostryalis.directory"))).expanduser()
        base_dir.mkdir(parents=True, exist_ok=True)
        return base_dir

    def _escape_sqlite_string_literal(self, value: str) -> str:
        # SQLite string literal escaping: '' inside '...'
        return value.replace("'", "''")

    def _get_attached_cache(self, connection_record) -> set[str]:
        info = connection_record.info
        cache = info.get(_ATTACHED_CACHE_KEY)
        if cache is None:
            cache = set()
            info[_ATTACHED_CACHE_KEY] = cache
        return cache

    def _get_udf_cache(self, connection_record) -> set[tuple[str, int]]:
        info = connection_record.info
        cache = info.get(_UDF_CACHE_KEY)
        if cache is None:
            cache = set()
            info[_UDF_CACHE_KEY] = cache
        return cache

    def _sqlite_install_udfs(
        self,
        dbapi_connection,
        *,
        required: Iterable[tuple[str, int]],
        connection_record=None,
    ) -> None:
        cache: set[tuple[str, int]] | None = None
        if connection_record is not None:
            cache = self._get_udf_cache(connection_record)
            required = [k for k in required if k not in cache]
            if not required:
                return

        # Install each function on this DB-API connection.
        for (name, arity) in required:
            with self._lock:
                func = self._udfs.get((name, arity))
            if func is None:
                raise KeyError(
                    f"Function '{name}/{arity}' is not registered; call register_function() first."
                )
            # sqlite3.Connection API
            dbapi_connection.create_function(name, arity, func)
            if cache is not None:
                cache.add((name, arity))

    def _sqlite_attach_aliases(
        self,
        dbapi_connection,
        *,
        aliases: Iterable[str],
        connection_record=None,
    ) -> None:
        cache: set[str] | None = None
        if connection_record is not None:
            cache = self._get_attached_cache(connection_record)
            aliases = [a for a in aliases if a not in cache]
            if not aliases:
                return

        cur = dbapi_connection.cursor()
        try:
            cur.execute("PRAGMA database_list")
            already = {row[1] for row in cur.fetchall()}

            for alias in aliases:
                if alias in already:
                    if cache is not None:
                        cache.add(alias)
                    continue

                with self._lock:
                    path = self._attachments.get(alias)
                if path is None:
                    # Defensive: don't KeyError; surface a clear error.
                    raise KeyError(f"Alias '{alias}' is not registered; call attach() first.")

                path_sql = self._escape_sqlite_string_literal(path)
                cur.execute(f"ATTACH DATABASE '{path_sql}' AS {alias}")

                if cache is not None:
                    cache.add(alias)
        finally:
            cur.close()

    def _sqlite_on_connect(self, dbapi_connection, connection_record) -> None:
        with self._lock:
            aliases = list(self._attachments.keys())
            udf_keys = list(self._udfs.keys())

        if aliases:
            self._sqlite_attach_aliases(
                dbapi_connection,
                aliases=aliases,
                connection_record=connection_record,
            )

        if udf_keys:
            self._sqlite_install_udfs(
                dbapi_connection,
                required=udf_keys,
                connection_record=connection_record,
            )

    def _get_session_factory(self) -> sessionmaker:
        if self._SessionLocal is None:
            self._get_engine()
        assert self._SessionLocal is not None
        return self._SessionLocal

    def _get_connection_record_from_sa_connection(self, sa_connection) -> object | None:
        try:
            fairy = sa_connection.connection  # _ConnectionFairy
            return getattr(fairy, "_connection_record", None)
        except Exception:
            return None

    def _ensure_required_attached(
        self,
        session: Session,
        require_attached: Iterable[str] | None,
    ) -> None:
        if not require_attached:
            return
        for alias in require_attached:
            self.attach(alias)

        sa_conn = session.connection()
        record = self._get_connection_record_from_sa_connection(sa_conn)
        self._sqlite_attach_aliases(
            sa_conn.connection,
            aliases=[self._validate_alias(a) for a in require_attached],
            connection_record=record,
        )

    def _ensure_required_functions(
        self,
        session: Session,
        require_functions: Iterable[tuple[str, int]] | None,
    ) -> None:
        if not require_functions:
            return

        normalized: list[tuple[str, int]] = [
            (self._validate_udf_name(n), self._validate_udf_arity(a))
            for (n, a) in require_functions
        ]

        sa_conn = session.connection()
        record = self._get_connection_record_from_sa_connection(sa_conn)
        self._sqlite_install_udfs(
            sa_conn.connection,
            required=normalized,
            connection_record=record,
        )

    # Test hooks (intended for unit tests; avoid in application code)

    @classmethod
    def reset_singleton(cls) -> None:
        with cls._singleton_lock:
            if cls._singleton is not None:
                try:
                    cls._singleton.reset()
                finally:
                    cls._singleton = None

    def set_config_getter(self, config_getter: Callable[[str], object]) -> None:
        with self._lock:
            self._config = config_getter
            # Changing config affects db location => must drop engine/pool.
            if self._engine is not None:
                self._engine.dispose()
            self._engine = None
            self._SessionLocal = None

    def reset(self) -> None:
        with self._lock:
            if self._engine is not None:
                self._engine.dispose()
            self._engine = None
            self._SessionLocal = None
            self._attachments.clear()
            self._udfs.clear()
